%% -------------------------------------------------------------------
%%
%% Copyright (c) 2010-2017 Basho Technologies, Inc.
%% Copyright (c) 2009-2010 Gemini Mobile Technologies, Inc.  All rights reserved.
%%
%% This file is provided to you under the Apache License,
%% Version 2.0 (the "License"); you may not use this file
%% except in compliance with the License.  You may obtain
%% a copy of the License at
%%
%%   http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing,
%% software distributed under the License is distributed on an
%% "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
%% KIND, either express or implied.  See the License for the
%% specific language governing permissions and limitations
%% under the License.
%%
%% -------------------------------------------------------------------

-module(cluster_info_basic).

%% Registration API
-export([register/0]).

%% Mandatory callbacks.
-export([cluster_info_init/0, cluster_info_generator_funs/0]).

%% Export in case anyone else thinks they might be useful.
-export([alarm_info/1, application_info/1,
         capture_ets_i/1, capture_net_kernel_i/1, nodes_info/1, capture_regs/1,
         erlang_memory/1, erlang_statistics/1,
         erlang_system_info/1, global_summary/1, inet_db_summary/1,
         loaded_modules/1, memory_hogs/2, non_zero_mailboxes/1, port_info/1,
         registered_names/1, time_and_date/1, timer_status/1]).

-ifdef(TEST).
-include_lib("eunit/include/eunit.hrl").
-endif.

register() ->
    cluster_info:register_app(?MODULE).

cluster_info_init() ->
    ok.

cluster_info_generator_funs() ->
    [
     %% Short(er) output items at the top
     {"Current time and date", time_and_date_fun()},
     {"VM statistics", fun erlang_statistics/1},
     {"erlang:memory() summary", fun erlang_memory/1},
     {"Top 50 process memory hogs", fun(C) -> memory_hogs(C, 50) end},
     {"Non-zero mailbox sizes", fun non_zero_mailboxes/1},
     {"Suspicious link and monitor counts", fun suspicious_links_monitors/1},
     {"Registered process names", fun registered_names/1},
     {"Registered process name via regs()", fun capture_regs/1},
     {"Ports", fun port_info/1},
     {"Applications", fun application_info/1},
     {"Timer status", fun timer_status/1},
     {"ETS summary", fun capture_ets_i/1},
     {"Nodes summary", fun nodes_info/1},
     {"net_kernel summary", fun capture_net_kernel_i/1},
     {"inet_db summary", fun inet_db_summary/1},
     {"Alarm summary", fun alarm_info/1},
     {"Global summary", fun global_summary/1},

     %% Longer output starts here.
     {"erlang:system_info() summary", fun erlang_system_info/1},
     {"Loaded modules", fun loaded_modules/1}
    ].

alarm_info(C) ->
    Alarms = gmt_util_get_alarms(),
    cluster_info:format(C, " Number of alarms: ~p\n", [length(Alarms)]),
    cluster_info:format(C, " ~p\n", [Alarms]).

application_info(C) ->
    cluster_info:format(C, " Application summary:\n"),
    cluster_info:format(C, " ~p\n\n", [application:info()]),
    [try
         cluster_info:format(C, " Application:get_all_key(~p):\n", [App]),
         cluster_info:format(C, " ~p\n\n", [application:get_all_key(App)]),
         cluster_info:format(C, " Application:get_all_env(~p):\n", [App]),
         cluster_info:format(C, " ~p\n\n", [application:get_all_env(App)])
     catch Class:Reason:Stacktrace ->
             cluster_info:format(C, "Error for ~p: ~p ~p at ~p\n",
                                 [App, Class, Reason, Stacktrace])
     end || App <- lists:sort([A || {A, _, _} <- application:which_applications()])].

capture_ets_i(C) ->
    %% Use the captured IO as the format string: it has \n embedded.
    cluster_info:format(
      C, cluster_info:capture_io(1000, fun() -> ets:i() end)).

capture_net_kernel_i(C) ->
    %% Use the captured IO as the format string: it has \n embedded.
    cluster_info:format(
      C, cluster_info:capture_io(1000, fun() -> net_kernel:i() end)).

capture_regs(C) ->
    %% Use the captured IO as the format string: it has \n embedded.
    cluster_info:format(
      C, cluster_info:capture_io(1000, fun() -> shell_default:regs() end)).

erlang_memory(C) ->
    cluster_info:format(C, " Native report:\n\n"),
    cluster_info:format(C, " ~p\n\n", [erlang:memory()]),
    cluster_info:format(C, " Report sorted by memory usage:\n\n"),
    SortFun = fun({_, X}, {_, Y}) -> X > Y end,
    cluster_info:format(C, " ~p\n", [lists:sort(SortFun, erlang:memory())]).

erlang_statistics(C) ->
    [cluster_info:format(C, " ~p: ~p\n", [Type, catch erlang:statistics(Type)])
     || Type <- [context_switches, exact_reductions, garbage_collection,
                 io, reductions, run_queue, runtime, wall_clock]].

erlang_system_info(C) ->
    Allocs = erlang:system_info(alloc_util_allocators),
    I0 = lists:flatten(
             [allocated_areas, allocator,
              [{allocator, Alloc} || Alloc <- Allocs],
              [{allocator_sizes, Alloc} || Alloc <- Allocs],
              c_compiler_used, check_io, compat_rel, cpu_topology,
              [{cpu_topology, X} || X <- [defined, detected]],
              creation, debug_compiled, dist_ctrl,
              driver_version, elib_malloc, fullsweep_after,
              garbage_collection, global_heaps_size, heap_sizes,
              heap_type, kernel_poll,
              logical_processors, machine, modified_timing_level,
              multi_scheduling, multi_scheduling_blockers,
              otp_release, process_count, process_limit,
              scheduler_bind_type, scheduler_bindings,
              scheduler_id, schedulers, schedulers_online,
              smp_support, system_version, system_architecture,
              threads, thread_pool_size, trace_control_word,
              version, wordsize]),
    _ = [cluster_info:format(C, " ~p:\n ~p\n\n", [I,catch erlang:system_info(I)]) ||
            I <- I0],

    I1 = [dist, info, loaded, procs],
    _ = [cluster_info:format(C, " ~p:\n ~s\n\n", [I, catch erlang:system_info(I)]) ||
            I <- I1].

global_summary(C) ->
    cluster_info:format(C, " info: ~p\n\n", [global:info()]),
    cluster_info:format(C, " registered_names:\n"),
    _ = [try
             Pid = global:whereis_name(Name),
             cluster_info:format(C, "    Name ~p, pid ~p, node ~p\n",
                                 [Name, Pid, node(Pid)])
         catch _:_ ->
                 ok
         end || Name <- global:registered_names()],
    ok.

inet_db_summary(C) ->
    cluster_info:format(C, " gethostname: ~p\n", [inet_db:gethostname()]),
    cluster_info:format(C, " get_rc: ~p\n", [inet_db:get_rc()]).

loaded_modules(C) ->
    cluster_info:format(C, " Root dir: ~p\n\n", [code:root_dir()]),
    cluster_info:format(C, " Lib dir: ~p\n\n", [code:lib_dir()]),
    %% io:format() output only {sniff!}
    %% cluster_info:format(C, " Clashes: ~p\n\n", [code:clash()]),
    PatchDirs = [Dir || Dir <- code:get_path(), string:str(Dir, "patch") /= 0],
    lists:foreach(
      fun(Dir) ->
              cluster_info:format(C, " Patch dir ~p:\n~s\n\n",
                                  [Dir, os:cmd("ls -l " ++ Dir)])
      end, PatchDirs),
    cluster_info:format(C, " All Paths:\n ~p\n\n", [code:get_path()]),
    [try
         cluster_info:format(C, " Module ~p:\n", [Mod]),
         cluster_info:format(C, " ~p\n\n", [Mod:module_info()])
     catch Class:Reason:Stacktrace ->
             cluster_info:format(C, "Error for ~p: ~p ~p at ~p\n",
                                 [Mod, Class, Reason, Stacktrace])
     end || Mod <- lists:sort([M || {M, _} <- code:all_loaded()])].

memory_hogs(C, Num) ->
    L = lists:sort([{Mem, Pid}
                    || Pid <- processes(),
                       {memory, Mem} <- [catch process_info(Pid, memory)],
                       is_integer(Mem)]),
    cluster_info:format(C, " ~p\n", [lists:sublist(lists:reverse(L), Num)]).

nodes_info(C) ->
    cluster_info:format(C, " My node: ~p\n", [node()]),
    cluster_info:format(C, " Other nodes: ~p\n", [lists:sort(nodes())]).

non_zero_mailboxes(C) ->
    L = lists:sort([{Size, Pid}
                    || Pid <- processes(),
                       {message_queue_len, Size} <-
                           [catch process_info(Pid, message_queue_len)],
                       Size > 0]),
    cluster_info:format(C, " ~p\n", [lists:reverse(L)]).

suspicious_links_monitors(C) ->
    Min = 1000,
    cluster_info:format(C, " Minimum threshold = ~p\n\n", [Min]),
    [begin
         L = lists:sort([{Num, Pid} ||
                            Pid <- processes(),
                            [{_, Mon}] <- [process_info(Pid, [Type])],
                            Num <- [length(Mon)],
                            Num >= Min]),
         cluster_info:format(C, " Type: ~p\n\n", [Type]),
         cluster_info:format(C, "    ~p\n\n", [lists:reverse(L)])
     end || Type <- [links, monitors]].

port_info(C) ->
    L = [{Port, catch erlang:port_info(Port)} || Port <- erlang:ports()],
    cluster_info:format(C, " ~p\n", [L]).

registered_names(C) ->
    L = lists:sort([{Name, catch whereis(Name)} || Name <- registered()]),
    cluster_info:format(C, " ~p\n", [L]).

time_and_date(C) ->
    case erlang:is_builtin(erlang, monotonic_time, 0) of
        true ->
            time_and_date_new(C);
        _ ->
            time_and_date_old(C)
    end.

time_and_date_fun() ->
    case erlang:is_builtin(erlang, monotonic_time, 0) of
        true ->
            fun time_and_date_new/1;
        _ ->
            fun time_and_date_old/1
    end.

time_and_date_old(C) ->
    cluster_info:format(C, " Current date: ~p\n", [erlang:date()]),
    cluster_info:format(C, " Current time: ~p\n", [erlang:time()]),
    cluster_info:format(C, " Current now : ~p\n", [os:timestamp()]).

time_and_date_new(C) ->
    cluster_info:format(C, " Current date     : ~p\n", [erlang:date()]),
    cluster_info:format(C, " Current time     : ~p\n", [erlang:time()]),
    cluster_info:format(C, " Current timestamp: ~p\n\n", [apply0(erlang, timestamp)]),
    cluster_info:format(C, " erlang:system_info(os_system_time_source):\n ~p\n\n",
        [erlang:system_info(os_system_time_source)]),
    cluster_info:format(C, " erlang:system_info(os_monotonic_time_source):\n ~p\n",
        [erlang:system_info(os_monotonic_time_source)]).

-compile({inline, apply0/2}).
apply0(Mod, Func) ->
    Mod:Func().

timer_status(C) ->
    case catch timer:get_status() of
        {'EXIT', _} ->
            cluster_info:format(C, " timer_server is not started");
        Status ->
            cluster_info:format(C, " ~p\n", [Status])
    end.

gmt_util_get_alarms() ->
    alarm_handler:get_alarms().

-ifdef(TEST).

time_and_date_test_() ->
    {timeout, 30, fun() ->
        This = erlang:self(),
        {Pid, Mon} = erlang:spawn_monitor(fun() -> time_and_date(This) end),
        Res = receive_time_and_date(Pid, Mon, []),
        ?assertMatch([_|_], Res),
        Len = case otp_version() of
            Vsn when Vsn >= 18 ->
                5;
            _ ->
                3
        end,
        ?assertEqual(Len, erlang:length(Res))
    end}.

receive_time_and_date(Pid, Mon, Acc) ->
    receive
        {collect_data, Pid, IoData} ->
            receive_time_and_date(Pid, Mon, [IoData | Acc]);
        {collect_data_ack, Pid} ->
            Pid ! collect_data_goahead,
            receive_time_and_date(Pid, Mon, Acc);
        {'DOWN', Mon, _, Pid, _} ->
            Acc
    after
        2000 ->
            {error, {timeout, Acc}}
    end.

otp_version() ->
    {Vsn, _} = string:to_integer(
        case erlang:system_info(otp_release) of
            [$R | Rel] ->
                Rel;
            Rel ->
                Rel
        end),
    Vsn.

-endif. % TEST
